/*
 * Data Integrity Test Tool
 *
 * This test is intended to verify data integrity after a system crash.
 * The idea is that there is a client which writes block data to disk
 * and notifies a server after the O/S signals the completion of each
 * write.  Then, via some external means, the box is crashed (pull the power
 * cord or issue a Sysrq-C perhaps).  When the system reboots, invoke
 * this program in check mode using the last data recorded by the server
 * to verify the on-disk contents.
 *
 * The client-server protocol is quite simple.  The client first sends
 * a generation number to the server.  It then proceeds to send the
 * block numbers after each block is written to disk.  When it wraps
 * from the end of the device back to the beginning, it increments the
 * generation number and starts over.
 *
 * Implementation Details
 *
 * On the client side, we use libaio to drive queue depths of the underlying
 * storage.  The maximum outstanding I/O is specified via a command line
 * parameter, and this is used to throttle I/O.  I/O is written sequentially
 * from block 0 to the last block of the device.  One thread runs the I/O
 * submission code, and a second thread reaps I/O completions and sends the
 * completed block numbers to the server.
 *
 * The server uses the maximum outstanding I/O number to size an array to
 * hold the values of the block numbers that are currently outstanding.
 * For example, if the maximum I/O depth is 32, then the server will create
 * an array (the iomap) with 32 entries in it.[1]  Each entry is primed with
 * the block number we are waiting for, which at the beginning will be blocks
 * 0 through 31.  The array looks like so:
 * [0,1,2,3,4,5,6,7,8,9,10,11,12...31]
 * When an I/O completion is reported, we look up the offset of the block
 * number into the array.  So, if block 3 completes, we look up index 3 in
 * the array, and check to ensure that this is the block we expect to complete.
 * We find that the value is 3, and we increment it by the array size.  The
 * new array looks like so:
 * [0,1,2,35,4,5,6,7,8,9,10,11,12..31]
 * The astute reader will notice that, if only a single I/O does not complete
 * before the other 31 I/Os in the array, we will run into block aliasing
 * problems.  This is what the IOMAP_MULTIPLIER is for.  We set this to a
 * number that is "reasonable."  For now, it is set at 2, and that seems
 * to work out well.  This means that, if the maximum outstanding I/O is
 * 32, we will size the array at 64 elements.
 *
 * Upon crash of the client system, you can issue a Ctrl-C on the server to
 * write out a file containing all of the data necessary to verify the disk
 * contents.  Using the array of outstanding I/O, we can derive which blocks
 * have completed.  The lowest-numbered array element will be 1 greater than
 * the last sequential I/O performed.  Using that number, we can determine
 * which I/Os were outstanding at the time of the crash.
 *
 * Torn pages will most likely occur on those still pending I/Os.  The
 * check option will walk through all of the sequential I/Os to ensure
 * that they have valid contents.  They will have the most up-to-date
 * generation number, the block number recorded in the disk block should
 * mention the block number in which they are located.  Finally, there is
 * a crc32 that should check out.
 *
 * After the sequential blocks are checked, we check the outstanding I/O
 * array with the previous generation number AND with the current generation
 * to see if they were either unwritten, or written but not acknowledged to
 * the O/S.
 *
 * Author: Jeff Moyer <jmoyer@redhat.com>
 *
 * Copyright 2007, 2008, Red Hat, Inc.
 * This program is licensed under the GNU GPL v2.
 */
/*
 *  TODO:
 *    - Clean up the debugging output on the server side.
 *    - print out offset within the block that corruption occurred.
 *    Automation Support:
 *    - Basically, make it so that the server drives the test.
 *    - add support for client configuration profiles that include:
 *      disk name / file name
 *      client IP address or hostname
 *      workload to run (10 runs of 32k, 10 runs of 1mb, etc)
 *    - report results to the server program
 *    - limit to a single trigger running at a time!
 */
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <features.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <netdb.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <limits.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/mount.h> /* required for BLKGETSIZE64 */
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <libaio.h>

/* defined in crc32.c */
uint32_t crc32(const char *data, size_t count);

const char *version = "0.99.3";

#ifndef BITS_PER_LONG
#ifdef _LP64
#define BITS_PER_LONG 64
#else
#define BITS_PER_LONG 32
#endif
#endif

#if BITS_PER_LONG == 64
#define UINT64_FMT "%lu"
#else
#define UINT64_FMT "%llu"
#endif

#ifdef DEBUG
#define dprintf(fmt, args...)	printf(fmt, ##args)
#else
#define dprintf(fmt, args...)
#endif

#define offsetof(TYPE, MEMBER)  (__builtin_offsetof(TYPE, MEMBER))

#define DEFAULT_BLOCK_SIZE	4096
#define DEFAULT_PORT		60606
#define DEVNAME_MAX_LEN		PATH_MAX
#define DEFAULT_IO_SUBMIT	16
#define DEFAULT_IO_DEPTH	64

/* the iomap size will be IOMAP_MULTIPLIER times the maximum I/O depth */
#define IOMAP_MULTIPLIER	2

struct block_data {
	uint32_t crc32;
	uint64_t generation_nr;
	uint64_t block_nr;
	char padding[0];
} __attribute__((__packed__));

/* client configuration data, also passed to the server via IPC */
struct config_data {
	char		dev[DEVNAME_MAX_LEN];
	uint32_t	block_size;
	uint32_t	max_io_depth;  /* max nr of outstanding ios */
	uint32_t	max_io_submit; /* nr ios per batch */
} __attribute__((packed));

/*
 * Data structure used by the server to keep track of the state of a client.
 */
struct client_state {
	struct config_data	config;
	uint64_t		generation_nr;
	unsigned		map_size;
	uint64_t		iomap[0];
} __attribute__((packed));
/* keep track of the previous state so clients can qeury it */
static struct client_state *last_state;

struct packet_header {
	uint32_t type;
#define BLOCK_PACKET		0x1
#define GENERATION_PACKET	0x2
#define CONFIG_PACKET		0x3
#define GETSTATE_PACKET		0x4 /* request the config from prevous run */
#define STATE_PACKET		0x5 /* send the state from a previous run */
	uint32_t len;
};

struct packet {
	struct packet_header header;
	union {
		uint64_t		block_nr;
		uint64_t		generation_nr;
		struct config_data	config;
		struct client_state	state;
	} un;
} __attribute__((packed));

/*
 * flag set from a signal handler to tell the server to stop
 * processing for a given client
 */
static int end_client = 0;

/*
 * Shared between the io submit thread and the io completion thread.
 */
struct writer_state {
	pthread_mutex_t mutex;
	pthread_cond_t  cond;
	int sock;
	io_context_t ioctx;
	unsigned long ios_in_flight;
	uint32_t max_io_depth;
} writer_state;

/* Currently, only one supported trigger type */
#define TRIGGER_TYPE_GENERATION	0
#define TRIGGER_TYPE_MAX	1
char *triggers[TRIGGER_TYPE_MAX];

void
copy_config(struct config_data *dst, struct config_data *src)
{
	*dst = *src;
}

void
init_iomap(uint64_t *iomap, unsigned buckets)
{
	unsigned i;

	for (i = 0; i < buckets; i++)
		iomap[i] = i;
}

struct client_state *
alloc_client_state(struct config_data *config, unsigned map_size)
{
	struct client_state *state;
	int state_size;

	/* If no map_size specified, calculate it using the MULTIPLIER */
	if (!map_size)
		map_size = config->max_io_depth * IOMAP_MULTIPLIER;
	state_size = (map_size * sizeof(uint64_t)) + sizeof(struct client_state);
	state = malloc(state_size);
	assert(state);
	memset(state, 0, state_size);

	copy_config(&state->config, config);
	state->generation_nr = (uint64_t)-1;
	state->map_size = map_size;
	init_iomap(state->iomap, state->map_size);

	return state;
}

/*
 *  Create an output file which contains all of the data stored in the
 *  client_state data structure.
 */
void
dump_state(struct client_state *state)
{
	int fd;
	unsigned i;
	FILE *fp;
	char template[] = "/tmp/client-XXXXXX";
	struct config_data *config = &state->config;

	fd = mkstemp(template);
	if (fd < 0) {
		perror("mkstemp");
		exit(1);
	}
	fp = fdopen(fd, "w");
	if (!fp) {
		perror("fdopen");
		exit(1);
	}

	fprintf(fp, "dev: %s\n", config->dev);
	fprintf(fp, "block_size: %u\n", config->block_size);
	fprintf(fp, "max_io_depth: %u\n", config->max_io_depth);
	fprintf(fp, "max_io_submit: %u\n", config->max_io_submit);
	fprintf(fp, "generation_nr: " UINT64_FMT "\n", state->generation_nr);
	fprintf(fp, "map_size: %u\n", state->map_size);
	for (i = 0; i < state->map_size; i++)
		fprintf(fp, "block: " UINT64_FMT "\n", state->iomap[i]);
	fflush(fp);
	fclose(fp);
	printf("Wrote %s.\n", template);
	return;
}

struct client_state *
read_state(char *data_file)
{
	FILE *fp;
	int i;
	int ret;
	struct config_data config;
	struct client_state *state;
	uint64_t block;
	uint64_t generation_nr;
	unsigned map_size;

	fp = fopen(data_file, "r");
	assert(fp);

	ret = fscanf(fp, "dev: %s\n", config.dev);
	assert(ret == 1);
	ret = fscanf(fp, "block_size: %u\n", &config.block_size);
	assert(ret == 1);
	ret = fscanf(fp, "max_io_depth: %u\n", &config.max_io_depth);
	assert(ret == 1);
	ret = fscanf(fp, "max_io_submit: %u\n", &config.max_io_submit);
	assert(ret == 1);
	ret = fscanf(fp, "generation_nr: " UINT64_FMT "\n", &generation_nr);
	assert(ret == 1);
	ret = fscanf(fp, "map_size: %u\n", &map_size);
	assert(ret == 1);

	state = alloc_client_state(&config, map_size);
	assert(state);
	state->generation_nr = generation_nr;
	state->map_size = map_size;

	i = 0;
	while (fscanf(fp, "block: " UINT64_FMT "\n", &block) != EOF)
		state->iomap[i++] = block;
	fclose(fp);

	return state;
}


int
open_dev(const char *block_device, uint64_t *device_size)
{
	int fd, ret;

	fd = open(block_device, O_RDWR | O_DIRECT | O_SYNC);
	if (fd < 0) {
		perror("open");
		exit(1);
	}

	/* Retrieve the device size in bytes. */
	ret = ioctl(fd, BLKGETSIZE64, device_size);
	if (ret < 0) {
		perror("ioctl: failed to get device size");
		exit(1);
	}

	return fd;
}

struct block_data *
alloc_block(size_t size)
{
	int ret;
	void *ptr;

	ret = posix_memalign(&ptr, getpagesize(), size);
	if (ret != 0) {
		errno = ret;
		perror("posix_memalign");
		return NULL;
	}

	return (struct block_data *)ptr;
}

void
fill_block(uint64_t generation_nr, uint64_t block_nr,
	   struct block_data *block, uint32_t block_size)
{
	unsigned char byte;

	assert(block_size >= sizeof(struct block_data));

	/*
	 * generation numbers start at 0.  We want to be able to initialize
	 * a file by doing a dd if=/dev/zero, so fill the device with 1's
	 * for the first generation, and 0's for the next.
	 */
	if (generation_nr % 2)
		byte = '0';
	else
		byte = '1';
	memset(block, byte, block_size);
	block->crc32 = 0;
	block->generation_nr = generation_nr;
	block->block_nr = block_nr;
	block->crc32 = crc32((const char *)block, block_size);
}

void
dump_header(struct block_data *block)
{
	printf("crc32: %u\n", block->crc32);
	printf("generation: " UINT64_FMT "\n", block->generation_nr);
	printf("block_nr: " UINT64_FMT "\n", block->block_nr);
}

void
dump_block(uint64_t block_nr, struct block_data *block, uint32_t block_size)
{
	int data_size = block_size - offsetof(struct block_data, padding);
	int header_size = block_size - data_size;
	int i, last_start = 0;
	int last = -1;
	int printed = 0;
	static int sizes_printed = 0;

	if (!sizes_printed) {
		printf("header size: %d, data size: %d\n",
		       header_size, data_size);
		sizes_printed = 1;
	}

	printf("Printing block contents for block " UINT64_FMT ".\n", block_nr);
	dump_header(block);
	for (i = 0; i < data_size; i++) {
		if (last == -1) {
			/* first time through */
			last = block->padding[i];
			last_start = i;
			continue;
		}

		if (block->padding[i] != last) {
			/* If things are changing block by block, then print
			 * out multiple blocks per line. */
			/* We add in header_size to get the offset in the block
			 * correct. */
			if (i - last_start == 1) {
				printf("%d: 0x%x\t", last_start + header_size,
				       last);
				printf("%d: 0x%x\t", i + header_size,
				       block->padding[i]);
				printed += 2;
				if (printed > 4) {
					printed = 0;
					printf("\n");
				}
			} else {
				if (printed) {
					printf("\n");
					printed = 0;
				}
				printf("%d - %d: 0x%x\n",
				       last_start + header_size,
				       i + header_size, last);
				last = block->padding[i];
				last_start = i;
			}
		}
	}

	if (!printed) /* the whole block was filled with the same stuff */
		printf("%d - %d: 0x%x\n", last_start + header_size,
		       i + header_size, last);
}

int
check_block(uint64_t generation_nr, uint64_t block_nr,
	    struct block_data *block, uint32_t block_size)
{
	uint32_t block_crc32;
	int ret = 0;

	assert(block_size >= sizeof(struct block_data));
	block_crc32 = block->crc32;
	block->crc32 = 0;

	if (block_crc32 != crc32((const char *)block, block_size)) {
		printf("DATA CORRUPTION: crcs do not match!\n");
		/* restore the original crc for printing */
		block->crc32 = block_crc32;
		dump_block(block_nr, block, block_size);
		ret = -1;
	}

	if (block_nr != block->block_nr) {
		printf("DATA CORRUPTION: block nr " UINT64_FMT " contains data "
		       "for block " UINT64_FMT "\n", block_nr, block->block_nr);
		ret = -1;
	}

	/*
	 * It is possible that the data was acknowledged by the O/S but the
	 * client did not send the packet to the server before the power
	 * loss/cable pull.
	 */
	if (block->generation_nr != generation_nr &&
	    block->generation_nr != generation_nr + 1) {
		printf("DATA CORRUPTION: While checking instance " UINT64_FMT
		       ", found data for instance " UINT64_FMT "\n",
		       generation_nr, block->generation_nr);
		ret = -1;
	}

	block->crc32 = block_crc32;
	return ret;
}

/*
 * TODO: optimize this to read in larger chunks of data.
 *       We should be able to use the same loop that writes the data to
 *       read it back.
 */
int
read_block(int fd, struct block_data *block,
	   unsigned block_size, uint64_t block_nr)
{
	ssize_t ret;
	char *buf = (char *)block;
	unsigned count = block_size;
	loff_t offset = block_nr * block_size;

retry:
	ret = pread(fd, buf, count, offset);
	if (ret < 0) {
		perror("pread");
		return -1;
	}

	if (ret == 0) {
		printf("unexpected EOF returned reading %u bytes from offset "
		       UINT64_FMT ".\n", count, offset);
		return -1;
	}

	if (ret < count) {
		count -= ret;
		offset += ret;
		buf += ret;
		goto retry;
	}

	return 0;
}

int
check_block_range(int fd, unsigned long start_block, unsigned long end_block,
		  uint64_t generation_nr, uint32_t block_size)
{
	uint64_t i;
	int ret = 0;
	struct block_data *block;

	block = alloc_block(block_size);
	if (!block)
		exit(1);

	dprintf("Checking generation " UINT64_FMT " from %ld to %ld\n",
		generation_nr, start_block, end_block);

	for (i = start_block; i <= end_block; i++) {
		if (read_block(fd, block, block_size, i) < 0) {
			printf("%s: start %lu end %lu generation " UINT64_FMT
			       " block_size %u, block " UINT64_FMT " failed.\n",
			       __func__, start_block, end_block, generation_nr,
			       block_size, i);
			exit(1);
		}
		/*
		 * First, check the crc and other fields inside
		 * the block for corruption.
		 */
		if (check_block(generation_nr, i, block, block_size) != 0) {
			ret = -1;
			printf("DATA CORRUPTION at block " UINT64_FMT ".\n", i);
		}
	}

	if (ret)
		printf("Instance " UINT64_FMT " failed check\n",
		       generation_nr);

	return ret;
}

int
recv_data(int sock, char *buf, size_t len)
{
	int ret, nready;
	fd_set rfds, xfds;
	struct timeval tv;

retry:
	FD_ZERO(&rfds);
	FD_ZERO(&xfds);
	FD_SET(sock, &rfds);
	tv.tv_sec = 15;
	tv.tv_usec = 0;

	nready = select(sock+1, &rfds, NULL, &xfds, &tv);
	if (nready <= 0) {
		if (nready < 0) {
			if (errno == EINTR)
				goto retry;
			perror("select");
		}
		/* timeout */
		return -1;
	}

	if (FD_ISSET(sock, &xfds)) {
		printf("exception on socket %d\n", sock);
		return -1;
	}

	ret = recv(sock, buf, len, MSG_DONTWAIT);
	if (ret == 0) {
		printf("Client closed connection\n");
		return -1;
	}

	if (ret != (ssize_t)len) {
		if (ret < 0) {
			if (errno == EINTR)
				goto retry;
			perror("recv");
			return -1;
		}
		buf = buf + ret;
		len -= ret;
		goto retry;
	}

	return 0;
}

struct packet *
recv_packet(int sock)
{
	int ret;
	struct packet_header header;
	struct packet *packet;

	ret = recv_data(sock, (char *)&header, sizeof(header));
	if (ret < 0)
		return NULL;

	packet = malloc(header.len);
	assert(packet);

	memcpy(packet, &header, sizeof(header));
	ret = recv_data(sock, (char *)&packet->un, header.len - sizeof(header));
	if (ret < 0) {
		free(packet);
		return NULL;
	}

	return packet;
}

void
send_packet(int sock, struct packet *packet)
{
	int ret;
	int count = packet->header.len;
	char *pos = (char *)packet;

	pthread_mutex_lock(&writer_state.mutex);
retry:
	ret = write(sock, pos, count);
	if (ret != count) {
		if (ret < 0) {
			if (errno == EINTR)
				goto retry;
			perror("write");
			exit(1);
		}
		count -= ret;
		pos += ret;
		goto retry;
	}
	pthread_mutex_unlock(&writer_state.mutex);
}

void
init_packet(struct packet *packet, int type, struct config_data *config,
	    struct client_state *state)
{
	memset(packet, 0, sizeof(*packet));
	packet->header.type = type;
	packet->header.len = sizeof(struct packet);
	if (type == CONFIG_PACKET) {
		assert(config);
		packet->un.config = *config;
	}
	if (type == STATE_PACKET) {
		assert(state);
		memset(&packet->un.state.iomap[0], 0,
		       state->map_size * sizeof(uint64_t));
		packet->header.len += (state->map_size * sizeof(uint64_t));
	}
}

/* TODO: this is not endianness clean */
void
send_config_packet(int sock, struct config_data *config)
{
	struct packet packet;

	init_packet(&packet, CONFIG_PACKET, config, NULL);
	send_packet(sock, &packet);
}

void
send_getstate_packet(int sock)
{
	struct packet packet;

	init_packet(&packet, GETSTATE_PACKET, NULL, NULL);
	send_packet(sock, &packet);
}

void
send_state_packet(int sock, struct client_state *state)
{
	struct packet *packet;
	unsigned i;

	packet = malloc(sizeof(*packet) + (state->map_size * sizeof(uint64_t)));
	assert(packet);
	init_packet(packet, STATE_PACKET, NULL, state);

	copy_config(&packet->un.state.config, &state->config);
	packet->un.state.generation_nr = state->generation_nr;
	packet->un.state.map_size = state->map_size;
	for (i = 0; i < state->map_size; i++) {
		dprintf("iomap[%u] = " UINT64_FMT "\n", i,
			state->iomap[i]);
		packet->un.state.iomap[i] = state->iomap[i];
	}

	send_packet(sock, packet);
}

void
send_block(int sock, uint64_t block_nr)
{
	struct packet packet;

	init_packet(&packet, BLOCK_PACKET, NULL, NULL);
	packet.un.block_nr = block_nr;
	send_packet(sock, &packet);
}

void
send_generation_nr(int sock, uint64_t generation_nr)
{
	struct packet packet;

	init_packet(&packet, GENERATION_PACKET, NULL, NULL);
	packet.un.generation_nr = generation_nr;
	send_packet(sock, &packet);
}

/*
 * Maybe extend to support udp or tcp, as other places do.
 */
int
connect_to_server(const char *servername, unsigned short port)
{
	int sock;
	struct hostent *hp;
	struct sockaddr_in servaddr;

	hp = gethostbyname(servername);
	if (!hp) {
		printf("DNS lookup of %s failed with %d\n", servername, errno);
		return -1;
	}

	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = PF_INET;
	servaddr.sin_port = htons(port);
	memcpy(&servaddr.sin_addr, hp->h_addr, sizeof(struct in_addr));

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("sock");
		return -1;
	}

	if (connect(sock, &servaddr, sizeof(servaddr)) != 0) {
		perror("connect");
		printf("Is the server actually running?\n");
		return -1;
	}

	return sock;
}

/*
 * Given the outstanding I/O map, determine whether a block has been written.
 * If the value in the array is the same as the block number, that means the
 * block is under I/O.  Otherwise, the value should be block number + map_size.
 */
int
block_completed(uint64_t block, struct client_state *state)
{
	int bucket = block % state->map_size;

	if (state->iomap[bucket] == block)
		return 0;
	assert(state->iomap[bucket] > block);
	return 1;
}

int
check_device(char *data_file, const char *server, unsigned short port)
{
	int fd, status, ret = 0;
	uint64_t i, start;
	uint64_t device_size;
	unsigned long end_block;

	struct config_data *config;
	struct client_state *state;
	uint64_t last_sequential_block;
	uint64_t min = ~0ULL;
	struct packet *packet = NULL;
	unsigned min_index = 0, num_written = 0;
	uint64_t window_start;

	if (data_file) {
		printf("checking file %s\n", data_file);
		/* Read in the data from the run */
		state = read_state(data_file);
	} else {
		int sock = connect_to_server(server, port);
		send_getstate_packet(sock);
		packet = recv_packet(sock);
		if (!packet) {
			printf("Unable to connect to server.  Try "
			       "specifying a data file\n");
			exit(1);
		}
		assert(packet->header.type == STATE_PACKET);
		state = &packet->un.state;
		close(sock);
	}
	config = &state->config;

	fd = open_dev(config->dev, &device_size);
	end_block = (device_size / config->block_size) - 1;
	printf("end_block: %lu\n", end_block);

	printf("state->map_size: %u\n", state->map_size);
	for (i = 0; i < state->map_size; i++) {
		dprintf("iomap[" UINT64_FMT "] = " UINT64_FMT "\n",
			i, state->iomap[i]);
		if (state->iomap[i] < min) {
			min = state->iomap[i];
			min_index = i;
		}
	}
	printf("min: " UINT64_FMT " min_index %u\n", min, min_index);

	/* The smallest numbered block is one greater than the last
	 * sequential completion */
	last_sequential_block = (min == ~0ULL) ? 0 : min - 1;
	printf("last_sequential_block: " UINT64_FMT "\n", last_sequential_block);

	/*
	 * The window then starts at the smallest block minus its offset
	 * in the array.
	 */
	window_start = min - min_index;

	/* Calculate how many of the entries in the window are completed */
	for (i = min_index + 1; i < config->max_io_depth; i++) {
		/*
		 *  If an array element is not window_start + offset, then
		 *  that I/O completed.
		 */
		if (state->iomap[i] != window_start + i)
			num_written++;
	}


	/*
	 * Given the last sequential I/O completion, the maximum I/O depth,
	 * and a list of completed I/Os, we can figure out which blocks
	 * should have been under I/O when the system crashed.
	 */
	start = last_sequential_block + 1;
	printf("start " UINT64_FMT " num_written %u\n", start, num_written);
	for (i = start; i < start + config->max_io_depth + num_written; i++) {
		/*
		 * We know that the block array never wraps to the beginning
		 * of the file without first flushing all I/O.  So, if we
		 * read to the end of the file, just reset i to 0.
		 */
		if (i > end_block) {
			i = 0;
			goto check_beginning;
		}

		/*
		 * If the block is under I/O, the contents are not
		 * guaranteed to be consistent.
		 */
		if (!block_completed(i, state))
			continue;

		status = check_block_range(fd, i, i, state->generation_nr,
					   config->block_size);
		if (status != 0)
			ret = -1;
	}

	/*
	 *  Check the remainder of the device.
	 */
	if (state->generation_nr != 0) {
		status = check_block_range(fd, i, end_block,
					   state->generation_nr - 1,
					   config->block_size);
		if (status != 0)
			ret = -1;
	}

check_beginning:
	/*
	 * Ensure that all of the I/O for this generation (up to the
	 * last sequential I/O) completed successfully.
	 */
	status = check_block_range(fd, 0, last_sequential_block,
				   state->generation_nr, config->block_size);
	if (status != 0)
		return -1;


	if (packet)
		free(packet);
	return ret;
}

void
sigint_handler(int __attribute__((unused)) signum)
{
	end_client = 1;
}

/*
 *  This function keeps tabs on a sliding window of packet completions.
 */
void
mark_written(struct client_state *state, unsigned long block_nr)
{
	int bucket = block_nr % state->map_size;

	if (state->iomap[bucket] != block_nr) {
		printf("got completion for I/O %lu before "
		       "completing I/O at " UINT64_FMT ".\n",
		       block_nr, state->iomap[bucket]);
		exit(1);
	}

	state->iomap[bucket] += state->map_size;
}

void
run_trigger(int type, int data)
{
	int ret;
	static pid_t child = 0;
	static int failed = 0;
	char trigger_arg[32];

	if (child) {
		int status;
		pid_t pid;
		pid = waitpid(child, &status, WNOHANG);
		if (pid == child) {
			if (WIFEXITED(status) && WEXITSTATUS(status))
				failed++;
		}
	}

	child = fork();
	switch(child) {
	case 0:
		break;
	case -1:
		printf("Failed to fork for generation trigger.\n");
		/* FALLTHROUGH */
	default:
		return;
	}

	snprintf(trigger_arg, sizeof(trigger_arg), "%d", data);
	ret = execl(triggers[type], triggers[type], trigger_arg, (char *)NULL);
	exit(1);
}

void
print_client_config(struct config_data *config)
{
	printf("New client:\n");
	printf("dev:\t%s\n", config->dev);
	printf("block size:\t%u\n", config->block_size);
	printf("max_io_depth:\t%u\n", config->max_io_depth);
	printf("max_io_submit:\t%u\n", config->max_io_submit);
}

/*
 * Server logic, called when a client connects.
 */
void
handle_client(int sock)
{
	struct packet *packet = NULL;
	struct sigaction sigint_action;
	struct client_state *state = NULL;

	/*
	 * A control-c from the terminal signals that we should dump
	 * the client configuration and close the connection.
	 */
	memset(&sigint_action, 0, sizeof(sigint_action));
	sigint_action.sa_handler = sigint_handler;
	sigint_action.sa_flags = SA_RESETHAND;
	if (sigaction(SIGINT, &sigint_action, NULL) < 0)
		perror("sigaction");

	while (1) {
		packet = recv_packet(sock);
		if (!packet)
			break;

		switch (packet->header.type) {
		case BLOCK_PACKET:
			assert(state);
			mark_written(state, packet->un.block_nr);
			break;
		case GENERATION_PACKET:
			assert(state);
			state->generation_nr = packet->un.generation_nr;
			/*
			 * Reset the iomap since we're starting
			 * I/O from the beginning of the device
			 * again.
			 */
			init_iomap(state->iomap, state->map_size);
			run_trigger(TRIGGER_TYPE_GENERATION,
				    state->generation_nr);
			break;
		case CONFIG_PACKET:
			state = alloc_client_state(&packet->un.config,0);
			print_client_config(&state->config);
			break;
		case GETSTATE_PACKET:
			if (!last_state)
				break;
			send_state_packet(sock, last_state);
			break;
		default:
			printf("Invalid packet type received: %d\n",
			       packet->header.type);
			return;
		}
		free(packet);
	}

	if (packet)
		free(packet);
	close(sock);

	/* dump the state to a file, and save a copy in case the client
	 * asks for it via GETSTATE. */
	if (state) {
		dump_state(state);
		if (last_state)
			free(last_state);
		last_state = state;
	}
}

/*
 * protocol can be either udp or tcp
 */
void
run_server(unsigned short port, const char *protocol)
{
	struct sockaddr_in servaddr, cliaddr;
	socklen_t clilen;
	int sock, sock_type, fd;
	int ret;
	int accept_errs = 0;

	if (strcmp(protocol, "udp") == 0) {
		sock_type = SOCK_DGRAM;
	} else if (strcmp(protocol, "tcp") == 0) {
		sock_type = SOCK_STREAM;
	} else {
		printf("Invalid protocol \"%s\".\n", protocol);
		exit(1);
	}

	memset(&servaddr, 0, sizeof(servaddr));

	servaddr.sin_family = PF_INET;
	servaddr.sin_port = htons(port);
	servaddr.sin_addr.s_addr = INADDR_ANY;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("socket");
		exit(1);
	}

	ret = bind(sock, (struct sockaddr *)&servaddr, sizeof(servaddr));
	if (ret < 0) {
		perror("bind");
		exit(1);
	}

	ret = listen(sock, 1);
	if (ret < 0) {
		perror("listen");
		exit(1);
	}

	while (1) {
		clilen = sizeof(cliaddr);
		fd = accept(sock, &cliaddr, &clilen);
		if (fd < 0) {
			perror("accept");
			if (++accept_errs > 100)
				exit(1);
			continue;
		}

		handle_client(fd);
		close(fd);
	}
}

void
inc_ios_in_flight(int nr_ios)
{
	assert(nr_ios > 0);
	pthread_mutex_lock(&writer_state.mutex);
	writer_state.ios_in_flight += nr_ios;
	pthread_mutex_unlock(&writer_state.mutex);
}

void
dec_ios_in_flight(int nr_ios)
{
	assert(nr_ios > 0);
	pthread_mutex_lock(&writer_state.mutex);
	writer_state.ios_in_flight -= nr_ios;
	pthread_cond_signal(&writer_state.cond);
	pthread_mutex_unlock(&writer_state.mutex);
}

void *
completion_thread(void  *arg)
{
	struct io_event *events, *event;
	io_context_t *ioctx = &writer_state.ioctx;
	/* the following is very suspicious, eh? */
	uint32_t max_io_depth = (uint32_t)((unsigned long)arg&(~0U));
	long ret;
	int i;

	events = malloc(max_io_depth * sizeof(struct io_event));
	if (!events) {
		perror("malloc");
		exit(1);
	}

	while (1) {
		ret = io_getevents(*ioctx, 1, max_io_depth, events, NULL);
		if (ret <= 0) {
			if (ret == -EINTR)
				continue;
			errno = -ret;
			perror("io_getevents");
			exit(1);
		}
		for (i = 0; i < ret; i++) {
			event = &events[i];
			send_block(writer_state.sock, *(uint64_t*)event->data);
			free(event->obj->u.c.buf);
			free(event->data);
			free(event->obj);
		}
		/* wakeup the thread issuing I/Os if it is waiting for
		 * free requests. */
		dec_ios_in_flight(ret);
	}

	return (void *)NULL;
}

void
submit_ios(io_context_t *ioctx, int nr, struct iocb **iocbs, uint32_t max_io_depth)
{
	int ret;

	assert((uint32_t)nr <= max_io_depth);
	/*
	 * Keep the number of outstanding writes within max_io_depth, and
	 * make sure there's enough room for the iocbs passed in.
	 */
	pthread_mutex_lock(&writer_state.mutex);
	while (writer_state.ios_in_flight + nr > max_io_depth)
		pthread_cond_wait(&writer_state.cond, &writer_state.mutex);
	pthread_mutex_unlock(&writer_state.mutex);

resubmit:
	ret = io_submit(*ioctx, nr, iocbs);
	if (ret != nr) {
		if (ret < 0) {
			if (ret == -EAGAIN)
				goto resubmit;
			errno = -ret;
			perror("io_submit");
			exit(1);
		}

		nr -= ret;
		iocbs = &iocbs[ret];
		inc_ios_in_flight(ret);
		goto resubmit;
	}
	inc_ios_in_flight(ret);
}

void
wait_for_all_ios(void)
{
	pthread_mutex_lock(&writer_state.mutex);
	while (writer_state.ios_in_flight)
		pthread_cond_wait(&writer_state.cond, &writer_state.mutex);
	pthread_mutex_unlock(&writer_state.mutex);
}

void
run_client(const char *servername, unsigned short port,
	   struct config_data *config)
{
	int fd, sock;
	uint64_t generation_nr = -1, i;
	uint64_t device_size;
	uint64_t blocks;
	struct block_data *block;
	struct iocb **iocbs, *iocb;
	io_context_t *ioctx = &writer_state.ioctx;
	pthread_t tid;
	long ret;

	memset(&writer_state, 0, sizeof(writer_state));
	pthread_mutex_init(&writer_state.mutex, NULL);
	pthread_cond_init(&writer_state.cond, NULL);
	writer_state.sock = -1;
	ret = io_setup(config->max_io_submit, &writer_state.ioctx);
	if (ret) {
		errno = -ret;
		perror("io_setup");
		exit(1);
	}

	ret = pthread_create(&tid, NULL, completion_thread,
			     (void *)(unsigned long)config->max_io_depth);
	if (ret) {
		perror("pthread_create");
		exit(1);
	}

	iocbs = malloc(config->max_io_submit * sizeof(struct iocb *));
	if (!iocbs) {
		perror("malloc");
		exit(1);
	}

	fd = open_dev(config->dev, &device_size);
	blocks = device_size / config->block_size;
	printf("di_client: writing " UINT64_FMT " blocks of size %u to "
	       "device %s, totalling " UINT64_FMT " bytes\n",
	       blocks, config->block_size, config->dev, device_size);
	printf("Submitting batches of %d ios, not to exceed %d ios in flight.\n",
		config->max_io_submit, config->max_io_depth);
	printf("Block Device Size: " UINT64_FMT ", reading " UINT64_FMT
	       "bytes.\n", device_size, blocks * config->block_size);

	sock = connect_to_server(servername, port);
	if (sock < 0)
		exit(1);
	writer_state.sock = sock;

	send_config_packet(sock, config);
	while (1) {
		generation_nr++;
		send_generation_nr(sock, generation_nr);

		printf("Writing data for generation number " UINT64_FMT "\n",
		       generation_nr);
		for (i = 0; i < blocks; i++) {
			/* Submit max_io_submit ios at a time. */
			if (i && i % config->max_io_submit == 0)
				submit_ios(ioctx, config->max_io_submit,
					   iocbs, config->max_io_depth);
			iocb = malloc(sizeof(struct iocb));
			block = alloc_block(config->block_size);
			assert(iocb);
			assert(block);

			fill_block(generation_nr, i, block, config->block_size);
			io_prep_pwrite(iocb, fd, block, config->block_size,
				       (uint64_t)(i*config->block_size));
			/* XXX not needed on 64 bit */
			iocb->data = malloc(sizeof(uint64_t));
			*(uint64_t *)iocb->data = i;
			iocbs[i % config->max_io_submit] = iocb;
		}

		/* If we hit the end of the device before max_io_submit,
		 * then submit what we have. */
		if (i % config->max_io_submit || i == blocks) {
			int nr = i % config->max_io_submit ?: config->max_io_submit;
			submit_ios(ioctx, nr, iocbs, config->max_io_depth);
		}

		wait_for_all_ios();
	}
}

int
client_prog(const char *prog)
{
	if (strstr(basename((char *)prog), "client"))
		return 1;
	return 0;
}

int
server_prog(const char *prog)
{
	if (strstr(basename((char *)prog), "server"))
		return 1;
	return 0;
}

void
register_trigger(int trigger_type, char *trigger)
{
	assert(trigger_type < TRIGGER_TYPE_MAX);
	triggers[trigger_type] = trigger;
}

void
usage(const char *prog)
{
	if (client_prog(prog))
		printf("Usage: %s [-c] [-v] -s <servername> -p <port> "
		       "-r <record_size> -d <io depth> -b <batch> "
		       "device_name\n", prog);
	else if (server_prog(prog))
		printf("Usage: %s -g <generation trigger> -p <port> [-v]\n",
		       prog);
	else
		printf("Binary should be named dainto_client or dainto_server.\n");

	exit(1);
}

#define CLIENT_OPTS	"hs:p:b:cr:d:v"
#define SERVER_OPTS	"hp:g:v"

int
main(int argc, char **argv)
{
	char *data_file = NULL;
	const char *option_string = NULL;
	char *servername = NULL;
	unsigned short port = DEFAULT_PORT;
	int c, ret;
	int check_mode = 0;
	struct config_data config;

	memset(&config, 0, sizeof(config));
	config.block_size = DEFAULT_BLOCK_SIZE;
	config.max_io_depth = DEFAULT_IO_DEPTH;
	config.max_io_submit = DEFAULT_IO_SUBMIT;

	if (client_prog(argv[0]))
		option_string = CLIENT_OPTS;
	else if (server_prog(argv[0]))
		option_string = SERVER_OPTS;
	else
		usage(argv[0]);

	while ((c = getopt(argc, argv, option_string)) != -1) {
		switch (c) {
		case 's':
			servername = strdup(optarg);
			if (!servername) {
				perror("strdup");
				exit(1);
			}
			break;

		case 'p':
			port = (unsigned short)atoi(optarg);
			break;

		case 'r': /* record size */
			config.block_size = (uint32_t)atoi(optarg);
			config.block_size *= 1024;
			break;

		case 'c':
			check_mode = 1;
			break;

		case 'd':
			config.max_io_depth = atoi(optarg);
			break;

		case 'b': /* batch */
			config.max_io_submit = atoi(optarg);
			break;

		case 'g': /* generation trigger */
			register_trigger(TRIGGER_TYPE_GENERATION, optarg);
			break;
		case 'v':
			printf("dainto version %s\n", version);
			exit(0);
		case '?':
		case 'h':
		default:
			usage(argv[0]);
			/* NOTREACHED */
		}
	}

	if (optind < argc) {
		if (argc - optind > 1)
			usage(argv[0]);
		if (check_mode)
			data_file = argv[optind];
		else
			strncat(config.dev, argv[optind], PATH_MAX - 1);
	}

	if (client_prog(argv[0])) {
		if (check_mode) {
			/* data_file may be null, in which case we check
			 * with the server for the data from the last run */
			if (!data_file && !servername)
				usage(argv[0]);
			ret = check_device(data_file, servername, port);
			exit(!!ret);
		} else {
			if (!*config.dev)
				usage(argv[0]);
			if (!servername) {
				printf("servername is required\n");
				usage(argv[0]);
			}

			run_client(servername, port, &config);
		}
	} else
		run_server(port, "tcp");

	exit(0);
}
